import numpy as np
from .activation import *

class lstm:
  def __init__(self, outshape, outactivation, hidshape=64, learningrate=0.05, inshape=None):
    self.outactivation, self.learningrate = outactivation, learningrate
    self.inshape, self.hidshape, self.outshape = inshape, hidshape, outshape
    
    if inshape is not None:
      self.weightsinToForgetGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
      self.weightsinToinGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
      self.weightsinTooutGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
      self.weightsinTocellMemGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))

    self.weightsHidToForgetGate = np.random.uniform(-0.005, 0.005, (hidshape, hidshape))
    self.weightsHidToinputGate = np.random.uniform(-0.005, 0.005, (hidshape, hidshape))
    self.weightsHidTooutputGate = np.random.uniform(-0.005, 0.005, (hidshape, hidshape))
    self.weightsHidTocellMemGate = np.random.uniform(-0.005, 0.005, (hidshape, hidshape))

    self.ForgetGatebiases = np.zeros(hidshape)
    self.inGatebiases = np.zeros(hidshape)
    self.outGatebiases = np.zeros(hidshape)
    self.cellMemGatebiases = np.zeros(hidshape)

    self.weightsHidToout = np.random.uniform(-0.005, 0.005, (hidshape, outshape))
    self.outbias = np.zeros(outshape)

  def __copy__(self):
    return type(self)(self.outshape, self.outactivation, self.hidshape, self.learningrate, self.inshape)

  def modelinit(self, inshape):
    self.inshape = inshape
    self.weightsinToForgetGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
    self.weightsinToinGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
    self.weightsinTooutGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
    self.weightsinTocellMemGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
    return self.outshape

  def forward(self, input):
    self.input = input
    self.cellSize = len(input)

    self.Hid = np.zeros((self.cellSize + 1, self.hidshape))
    self.cellMem = np.zeros((self.cellSize + 1, self.hidshape))
    self.out = np.zeros((self.cellSize, self.outshape))

    self.ForgetGate = np.zeros((self.cellSize, self.hidshape))
    self.inGate = np.zeros((self.cellSize, self.hidshape))
    self.outGate = np.zeros((self.cellSize, self.hidshape))
    self.cellMemGate = np.zeros((self.cellSize, self.hidshape))

    for i in range(self.cellSize):
      self.ForgetGate[i, :] = sigmoid().forward((self.weightsinToForgetGate.T @ self.input[i, :]) + (self.weightsHidToForgetGate.T @ self.Hid[i, :]) + self.ForgetGatebiases)
      self.inGate[i, :] = sigmoid().forward((self.weightsinToinGate.T @ self.input[i, :]) + (self.weightsHidToinputGate.T @ self.Hid[i, :]) + self.inGatebiases)
      self.outGate[i, :] = sigmoid().forward((self.weightsinTooutGate.T @ self.input[i, :]) + (self.weightsHidTooutputGate.T @ self.Hid[i, :]) + self.outGatebiases)
      self.cellMemGate[i, :] = tanh().forward((self.weightsinTocellMemGate.T @ self.input[i, :]) + (self.weightsHidTocellMemGate.T @ self.Hid[i, :]) + self.cellMemGatebiases)

      self.cellMem[i + 1, :] = (self.ForgetGate[i, :] * self.cellMem[i, :]) + (self.inGate[i, :] * self.cellMemGate[i, :])
      self.Hid[i + 1, :] = self.outGate[i, :] * tanh().forward(self.cellMem[i + 1, :])
      self.out[i, :] = self.outactivation.forward(self.weightsHidToout.T @ self.Hid[i + 1, :] + self.outbias)

    return self.out

  def backward(self, outError):
    inError = np.zeros(self.input.shape)
    HidError = np.zeros(self.hidshape)
    cellMemError = np.zeros(self.hidshape)

    weightsinToForgetGateΔ = np.zeros(self.weightsinToForgetGate.shape)
    weightsinToinGateΔ = np.zeros(self.weightsinToinGate.shape)
    weightsinTooutGateΔ = np.zeros(self.weightsinTooutGate.shape)
    weightsinTocellMemGateΔ = np.zeros(self.weightsinTocellMemGate.shape)

    weightsHidToForgetGateΔ = np.zeros(self.weightsHidToForgetGate.shape)
    weightsHidToinputGateΔ = np.zeros(self.weightsHidToinputGate.shape)
    weightsHidTooutputGateΔ = np.zeros(self.weightsHidTooutputGate.shape)
    weightsHidTocellMemGateΔ = np.zeros(self.weightsHidTocellMemGate.shape)

    ForgetGatebiasesΔ = np.zeros(self.ForgetGatebiases.shape)
    inGatebiasesΔ = np.zeros(self.inGatebiases.shape)
    outGatebiasesΔ = np.zeros(self.outGatebiases.shape)
    cellMemGatebiasesΔ = np.zeros(self.cellMemGatebiases.shape)

    weightsHidTooutΔ = np.zeros(self.weightsHidToout.shape)
    outbiasΔ = np.zeros(self.outbias.shape)

    for i in reversed(range(self.cellSize)):
      outGradient = self.outactivation.backward(self.out[i, :]) * outError[i, :]
      
      HidError += self.weightsHidToout @ outError[i, :]

      cellMemError += HidError * self.outGate[i] * tanh().backward(tanh().forward(self.cellMem[i + 1, :]))
      outGateError = HidError * tanh().forward(self.cellMem[i + 1, :])

      ForgetGateError = cellMemError * self.cellMem[i, :]
      inGateError = cellMemError * self.cellMemGate[i, :]
      cellMemGateError = cellMemError * self.inGate[i, :]

      cellMemError *= self.ForgetGate[i, :]

      ForgetGateGradient = sigmoid().backward(self.ForgetGate[i, :]) * ForgetGateError
      inGateGradient = sigmoid().backward(self.inGate[i, :]) * inGateError
      outGateGradient = sigmoid().backward(self.outGate[i, :]) * outGateError
      cellMemGateGradient = tanh().backward(self.cellMemGate[i, :]) * cellMemGateError

      HidError = self.weightsHidToForgetGate @ ForgetGateError + self.weightsHidToinputGate @ inGateError + self.weightsHidTooutputGate @ outGateError + self.weightsHidTocellMemGate @ cellMemGateError
      inError[i, :] = self.weightsinToForgetGate @ ForgetGateError + self.weightsinToinGate @ inGateError + self.weightsinTooutGate @ outGateError + self.weightsinTocellMemGate @ cellMemGateError

      weightsinToForgetGateΔ += np.outer(self.input[i, :].T, ForgetGateGradient)
      weightsinToinGateΔ += np.outer(self.input[i, :].T, inGateGradient)
      weightsinTooutGateΔ += np.outer(self.input[i, :].T, outGateGradient)
      weightsinTocellMemGateΔ += np.outer(self.input[i, :].T, cellMemGateGradient)

      weightsHidToForgetGateΔ += np.outer(self.Hid[i, :].T, ForgetGateGradient)
      weightsHidToinputGateΔ += np.outer(self.Hid[i, :].T, inGateGradient)
      weightsHidTooutputGateΔ += np.outer(self.Hid[i, :].T, outGateGradient)
      weightsHidTocellMemGateΔ += np.outer(self.Hid[i, :].T, cellMemGateGradient)

      ForgetGatebiasesΔ += ForgetGateGradient
      inGatebiasesΔ += inGateGradient
      outGatebiasesΔ += outGateGradient
      cellMemGatebiasesΔ += cellMemGateGradient

      weightsHidTooutΔ += np.outer(self.Hid[i].T, outGradient)
      outbiasΔ += outGradient

    self.weightsinToForgetGate += np.clip(weightsinToForgetGateΔ, -1, 1) * self.learningrate
    self.weightsinToinGate += np.clip(weightsinToinGateΔ, -1, 1) * self.learningrate
    self.weightsinTooutGate += np.clip(weightsinTooutGateΔ, -1, 1) * self.learningrate
    self.weightsinTocellMemGate += np.clip(weightsinTocellMemGateΔ, -1, 1) * self.learningrate

    self.weightsHidToForgetGate += np.clip(weightsHidToForgetGateΔ, -1, 1) * self.learningrate
    self.weightsHidToinputGate += np.clip(weightsHidToinputGateΔ, -1, 1) * self.learningrate
    self.weightsHidTooutputGate += np.clip(weightsHidTooutputGateΔ, -1, 1) * self.learningrate
    self.weightsHidTocellMemGate += np.clip(weightsHidTocellMemGateΔ, -1, 1) * self.learningrate

    self.ForgetGatebiases += np.clip(ForgetGatebiasesΔ, -1, 1) * self.learningrate
    self.inGatebiases += np.clip(inGatebiasesΔ, -1, 1) * self.learningrate
    self.outGatebiases += np.clip(outGatebiasesΔ, -1, 1) * self.learningrate
    self.cellMemGatebiases += np.clip(cellMemGatebiasesΔ, -1, 1) * self.learningrate

    self.weightsHidToout += np.clip(weightsHidTooutΔ, -1, 1) * self.learningrate
    self.outbias += np.clip(outbiasΔ, -1, 1) * self.learningrate

    return inError


class texttobinary:
  def forward(input):
    Dec = list(bytearray(input, "utf8"))
    out = [''] * len(input)
    for i in range(len(input)):
      out[i] = bin(Dec[i]).replace("b", ("0"*(9-len(bin(Dec[i])))))
    return out
  def backward(input):
    return input

class binarytotext:
  def forward(input):
    out = ''
    for i in range(len(input)):
      out += chr(int(input[i], 2))
    return out
  def backward(input):
    return input


import numpy as np
from .activation import *

class rnn:
  def __init__(self, outshape, Type, outactivation=stablesoftmax(), hidshape=64, learningrate=0.05, inshape=None):
    self.learningrate, self.Type, self.outactivation = learningrate, Type, outactivation
    self.inshape, self.hidshape, self.outshape = inshape, hidshape, outshape
    
    if inshape is not None:
      self.weightsinToHid = np.random.uniform(-0.005, 0.005, (hidshape, inshape))

    self.weightsHidToHid = np.random.uniform(-0.005, 0.005, (hidshape, hidshape))
    self.Hidbiases = np.zeros(hidshape)

    self.weightsHidToout = np.random.uniform(-0.005, 0.005, (outshape, hidshape))
    self.outbiases = np.zeros(outshape)

  def __copy__(self):
    return type(self)(self.outshape, self.Type, self.outactivation, self.hidshape, self.learningrate, self.inshape)

  def modelinit(self, inshape):
    self.inshape = inshape
    self.weightsinToHid = np.random.uniform(-0.005, 0.005, (hidshape, inshape))
    return self.outshape

  def forward(self, input):
    self.input = input
    self.Hid = np.zeros((len(self.input) + 1, self.hidshape))
    
    if self.Type == 'ManyToOne':
      for i in range(len(input)):
        self.Hid[i + 1, :] = tanh().forward(self.weightsinToHid @ input[i] + self.weightsHidToHid @ self.Hid[i, :] + self.Hidbiases)

      self.out = self.outactivation.forward(self.weightsHidToout @ self.Hid[len(input), :] + self.outbiases)
    
    elif self.Type == 'ManyToMany':
      self.out = np.zeros((len(self.input), self.outshape))

      for i in range(len(input)):
        self.Hid[i + 1, :] = tanh().forward(self.weightsinToHid @ input[i] + self.weightsHidToHid @ self.Hid[i, :] + self.Hidbiases)
        self.out[i, :] = self.outactivation.forward(self.weightsHidToout @ self.Hid[i + 1, :] + self.outbiases)

    return self.out

  def backward(self, outError):
    weightsinToHidΔ = np.zeros(self.weightsinToHid.shape)
    weightsHidToHidΔ = np.zeros(self.weightsHidToHid.shape)
    HidbiasesΔ = np.zeros(self.Hidbiases.shape)

    if self.Type == 'ManyToOne':
      outGradient = np.multiply(self.outactivation.backward(self.out), outError)

      weightsHidTooutΔ = np.outer(outGradient, self.Hid[len(self.input)].T)
      outbiasesΔ = outGradient

      HidError = self.weightsHidToout.T @ outError

      for i in reversed(range(len(self.input))):
        HidGradient = np.multiply(tanh().backward(self.Hid[i + 1]), HidError)

        HidbiasesΔ += HidGradient
        weightsHidToHidΔ += np.outer(HidGradient, self.Hid[i].T)
        weightsinToHidΔ += np.outer(HidGradient, self.input[i].T)

        HidError = self.weightsHidToHid.T @ HidGradient

    elif self.Type == 'ManyToMany':
      weightsHidTooutΔ = np.zeros(self.weightsHidToout.shape)
      outbiasesΔ = np.zeros(self.outbiases.shape)

      HidError = self.weightsHidToout.T @ outError[len(self.input) - 1]

      for i in reversed(range(len(self.input))):
        HidGradient = np.multiply(tanh().backward(self.Hid[i + 1]), HidError)
        outGradient = np.multiply(self.outactivation.backward(self.out[i]), outError[i])

        weightsinToHidΔ += np.outer(HidGradient, self.input[i].T)
        weightsHidToHidΔ += np.outer(HidGradient, self.Hid[i].T)
        HidbiasesΔ += HidGradient

        weightsHidTooutΔ += np.outer(outGradient, self.Hid[i].T)
        outbiasesΔ += outGradient

        HidError = self.weightsHidToHid.T @ HidGradient + self.weightsHidToout.T @ outError[i]

    self.weightsinToHid += self.learningrate * np.clip(weightsinToHidΔ, -1, 1)
    self.weightsHidToHid += self.learningrate * np.clip(weightsHidToHidΔ, -1, 1)
    self.Hidbiases += self.learningrate * np.clip(HidbiasesΔ, -1, 1)

    self.weightsHidToout += self.learningrate * np.clip(weightsHidTooutΔ, -1, 1)
    self.outbiases += self.learningrate * np.clip(outbiasesΔ, -1, 1)


import numpy as np

'''
- add GeLU
- add ELU
- add PReLU
'''

class sigmoid:
  def forward(self, input):
    return 1 / (1 + np.exp(-input))
  def backward(self, input):
    return input * (1 - input)

class tanh:
  def forward(self, input):
    return np.tanh(input)
  def backward(self, input):
    return 1 - np.square(input)

class relu:
  def forward(self, input):
    output = np.maximum(0, input)
    return output
  def backward(self, input):
    equation = np.vectorize(self.equationforderivative, otypes=[float])
    return equation(input)
  def equationforderivative(self, input):
    return 0 if (input <= 0) else 1

class leakyrelu:
  def __init__(self, alpha=0.01):
    self.alpha = alpha
  def forward(self, input):
    return np.maximum(self.alpha * input, input)
  def backward(self, input):
    equation = np.vectorize(self.equationforderivative, otypes=[float])
    return equation(input)
  def equationforderivative(self, input):
    return self.alpha if input < 0 else 1

class identity:
  def forward(self, input):
    return input
  def backward(self, input):
    return 1
  
class binarystep:
  def forward(self, input):
    equation = np.vectorize(self.equationForbinaryStep, otypes=[float])
    return equation(input)
  def equationForbinaryStep(self, input):
    return 0 if (input < 0) else 1
  def backward(self, input):
    return 1
    
class selu:
  def forward(self, input):
    equation = np.vectorize(self.equationForSELU, otypes=[float])
    return 1.0507 * equation(input)
  def equationForSELU(self, input):
    return 1.67326 * (np.exp(input) - 1) if (input < 0) else input
  def backward(self, input):
    equation = np.vectorize(self.equationforderivative, otypes=[float])
    return 1.0507 * equation(input)
  def equationforderivative(self, input):
    return 1.67326 * np.exp(input) if (input < 0) else 1

class silu:
  def forward(self, input):
    return input / (1 + np.exp(-input))
  def backward(self, input):
    return (1 + np.exp(-input) + (input * np.exp(-input))) / np.square(1 + np.exp(-input))
  
class mish:
  def forward(self, input):
    return input * Tanh().Tanh(np.log(1 + np.exp(input)))
  def backward(self, input):
    return (np.exp(input) * ((4 * np.exp(2 * input)) + np.exp(3 * input) + (4 * (1 + input)) + (np.exp(input) * (6 + (4 * input))))) / np.square(2 + (2 * np.exp(input)) + np.exp(2 * input))

class gaussian:
  def forward(self, input):
    return np.exp(-np.square(input))
  def backward(self, input):
    return -2 * input * np.exp(-np.square(input))

class softplus:
  def forward(self, input):
    return np.log(1 + np.exp(input))
  def backward(self, input):
    return 1 / (1 + np.exp(-input))

class softmax:
  def forward(self, input):
    return (input - np.max(input)) / np.sum(input - np.max(input)) # Check
  def backward(self, input):
    equation = np.vectorize(self.equationforderivative, otypes=[float])
    return equation(input, np.sum(np.exp(input)))
  def equationforderivative(self, input, SumExpOfinput):
    return (np.exp(input) * (SumExpOfinput - np.exp(input))) / (SumExpOfinput) ** 2

class stablesoftmax:
  def forward(self, input):
    return np.exp(input - np.max(input)) / np.sum(np.exp(input - np.max(input)))
  def backward(self, input):
    equation = np.vectorize(self.equationforderivative, otypes=[float])
    return equation(input, np.sum(np.exp(input)))
  def equationforderivative(self, input, SumExpOfinput):
    return (np.exp(input) * (SumExpOfinput - np.exp(input))) / (SumExpOfinput) ** 2


import numpy as np

class dropout:
  def __init__(self, dropoutrate):
    self.dropoutrate = dropoutrate

  def __copy__(self):
    return type(self)(self.dropoutrate)

  def modelinit(self, inshape):
    return inshape

  def forward(self, input):
    self.Dropout = np.random.binomial(1, self.dropoutrate, size=input.shape)
    self.Dropout = np.where(self.Dropout == 0, 1, 0)
    return self.Dropout * input

  def backward(self, outError):
    return self.Dropout * outError  
    
  def changeDropoutRate(self, NewRate):
    self.dropoutrate = NewRate


import numpy as np
from .activation import *

class nn:
  def __init__(self, outshape, activation, learningrate, weightsinit=(-1, 1), biasesinit=(0, 0), inshape=None):
    self.weightsinit, self.biasesinit = weightsinit, biasesinit
    self.activation, self.learningrate = activation, learningrate
    self.inshape = inshape
    if inshape is not None:
      self.weights = np.random.uniform(weightsinit[0], weightsinit[1], (np.prod(inshape), np.prod(outshape)))
      self.biases = np.random.uniform(biasesinit[0], biasesinit[1], np.prod(outshape))
    self.outshape = outshape
    
  def __copy__(self):
    return type(self)(self.outshape, self.activation, self.learningrate, self.weightsinit, self.biasesinit, self.inshape)

  def modelinit(self, inshape):
    self.inshape = inshape

    self.weights = np.random.uniform(self.weightsinit[0], self.weightsinit[1], (np.prod(inshape), np.prod(self.outshape)))
    self.biases = np.random.uniform(self.biasesinit[0], self.biasesinit[1], np.prod(self.outshape))
    return self.outshape

  def forward(self, input):
    self.input = input.flatten()
    self.out = self.weights.T @ self.input + self.biases
    
    self.out = self.activation.forward(self.out)

    return self.out.reshape(self.outshape)

  def backward(self, outError):
    outError = outError.flatten()
    
    outGradient = self.activation.backward(self.out) * outError
      
    inputError = self.weights @ outError
      
    self.biases += outGradient * self.learningrate
    self.weights += np.outer(self.input.T, outGradient) * self.learningrate
    return inputError.reshape(self.inshape)


import numpy as np
from alive_progress import alive_bar
from copy import copy

class neuroevolution:
  def __init__(self, inshape, outshape, PopulationSize, model):
    self.inshape, self.outshape, self.PopulationSize = inshape, outshape, PopulationSize
    
    self.model = np.zeros((PopulationSize, len(model)), dtype=object)
    for i in range(PopulationSize):
      inshape = self.inshape
      for j in range(len(model)):
        self.model[i, j] = copy(model[j])
        inshape = self.model[i, j].modelinit(inshape)

  def forwardmulti(self, input):
    out = np.zeros((self.PopulationSize, self.outshape))
    for i in range(self.PopulationSize):
      Hid = input
      for j in range(self.model.shape[1]):
        Hid = self.model[i, j].forward(Hid)
      out[i] = Hid
    return out

  def forwardsingle(self, input, Player):
    out = np.zeros(self.outshape)
    for j in range(self.model.shape[1]):
      input = self.model[Player, j].forward(input)
    return input

  def mutate(self, FavorablePlayer):
    Favorablemodel = self.model[FavorablePlayer]

    for i in range(self.model.shape[1]):
      self.model[0, i].weights = Favorablemodel[i].weights
      self.model[0, i].biases = Favorablemodel[i].biases

    for i in range(1, self.PopulationSize):
      for j in range(self.model.shape[1]):
        self.model[i, j].weights = Favorablemodel[j].weights
        self.model[i, j].biases = Favorablemodel[j].biases

        self.model[i, j].weights *= np.random.choice([0, 1], self.model[i, j].weights.shape, p=[self.model[i, j].learningrate, 1 - self.model[i, j].learningrate])
        self.model[i, j].weights = np.where(self.model[i, j].weights == 0, np.random.uniform(-1, 1, self.model[i, j].weights.shape), self.model[i, j].weights)

        self.model[i, j].biases *= np.random.choice([0, 1], self.model[i, j].biases.shape, p=[self.model[i, j].learningrate, 1 - self.model[i, j].learningrate])
        self.model[i, j].biases = np.where(self.model[i, j].biases == 0, np.random.uniform(-1, 1, self.model[i, j].biases.shape), self.model[i, j].biases)


from .activation import *
from .conv import conv
from .dropout import dropout
from .gru import gru
from .lstm import lstm
from .model import model
from .neuroevolution import neuroevolution
from .nn import nn
from .pool import pool
from .rnn import rnn
from .texttobinary import texttobinary


import numpy as np
from .activation import *

class conv:
  def __init__(self, filtershape, learningrate, activation=identity, padding=False, stride=(1, 1), inshape=None):
    self.learningrate, self.activation, self.padding, self.stride = learningrate, activation, padding, stride
    self.inshape = inshape

    self.filtershape = filtershape
    self.Filter = np.random.uniform(-0.25, 0.25, (self.filtershape))
    self.bias = np.zeros(self.filtershape[0])

    if inshape is not None:
      if len(inshape) == 2:
        inshape = tuple([self.filtershape[0]]) + inshape
      if padding == True:
        inshape = (inshape[0], inshape[1] + self.filtershape[1] - 1, inshape[2] + self.filtershape[2] - 1)
      self.outshape = tuple([filtershape[0], int((inshape[1] - filtershape[1] + 1) / self.stride[0]), int((inshape[2] - filtershape[2] + 1) / self.stride[1])])
      self.out = np.zeros(self.outshape)

  def __copy__(self):
    return type(self)(self.filtershape, self.learningrate, self.activation, self.padding, self.stride, self.inshape)

  def modelinit(self, inshape):
    if len(inshape) == 2:
      inshape = tuple([self.filtershape[0]]) + inshape
    if self.padding == True:
      inshape = (inshape[0], inshape[1] + self.filtershape[1] - 1, inshape[2] + self.filtershape[2] - 1)
    self.outshape = tuple([self.filtershape[0], int((inshape[1] - self.filtershape[1] + 1) / self.stride[0]), int((inshape[2] - self.filtershape[2] + 1) / self.stride[1])])
    self.out = np.zeros(self.outshape)

    return self.outshape

  def SetSlopeForLeakyReLU(self, Slope):
    LeakyReLU.Slope = Slope

  def forward(self, input):
    self.input = input
    if(input.ndim == 2):
      self.input = np.stack(([self.input] * self.filtershape[0]))
    if (self.padding == True):
      self.input = np.pad(self.input, int((len(self.Filter[0]) - 1) / 2), mode='constant')[1 : self.filtershape[0] + 1]

    for i in range(0, self.outshape[1], self.stride[0]):
      for j in range(0, self.outshape[2], self.stride[1]):
        self.out[:, i, j] = np.sum(np.multiply(self.input[:, i : i + self.filtershape[1], j : j + self.filtershape[2]], self.Filter), axis=(1, 2))

    self.out += self.bias[:, np.newaxis, np.newaxis]
    self.out = self.activation.forward(self.out)

    return self.out
  
  def backward(self, outError):
    FilterΔ = np.zeros(self.filtershape)
    
    outGradient = self.activation.backward(self.out) * outError

    for i in range(0, self.outshape[1], self.stride[0]):
      for j in range(0, self.outshape[2], self.stride[1]):
        FilterΔ += self.input[:, i : i + self.filtershape[1], j : j + self.filtershape[2]] * outGradient[:, i, j][:, np.newaxis, np.newaxis]
    
    self.bias += np.sum(outGradient, axis=(1, 2)) * self.learningrate
    self.Filter += FilterΔ * self.learningrate

    # in Error
    RotFilter = np.rot90(np.rot90(self.Filter))
    PaddedError = np.pad(outError, self.filtershape[1] - 1, mode='constant')[self.filtershape[1] - 1 : self.filtershape[0] + self.filtershape[1] - 1, :, :]
    
    self.inError = np.zeros(self.inshape)
    if np.ndim(self.inError) == 3:
      for i in range(int(self.inshape[1] / self.stride[0])):
        for j in range(int(self.inshape[2] / self.stride[1])):
         self.inError[:, i * self.stride[0], j * self.stride[1]] = np.sum(np.multiply(RotFilter, PaddedError[:, i:i + self.filtershape[1], j:j + self.filtershape[2]]), axis=(1, 2))
       
    if np.ndim(self.inError) == 2:
      for i in range(int(self.inshape[0] / self.stride[0])):
        for j in range(int(self.inshape[1] / self.stride[1])):
         self.inError[i * self.stride[0], j * self.stride[1]] = np.sum(np.multiply(RotFilter, PaddedError[:, i:i + self.filtershape[1], j:j + self.filtershape[2]]))

    return self.inError


from setuptools import setup

setup(name='aiinpy',
      version='0.0.17',
      description='An artificial intelligence library?',
      license='MIT',
      packages=['aiinpy'],
      url='https://github.com/SeanMabli/aiinpy',
      install_requires=['numpy', 'alive_progress'])


import numpy as np
from alive_progress import alive_bar

class model:
  def __init__(self, inshape, outshape, model):
    self.inshape = inshape = inshape if isinstance(inshape, tuple) else tuple([inshape])
    self.outshape = outshape if isinstance(outshape, tuple) else tuple([outshape])
    self.model = model

    for i in self.model:
      inshape = i.modelinit(inshape)

  def forward(self, input):
    for i in range(len(self.model)):
      input = self.model[i].forward(input)
    return input

  def backward(self, outError):
    for i in reversed(range(len(self.model))):
      outError = self.model[i].backward(outError)
    return outError

  def train(self, data, NumOfGen):
    # data preprocessing: tuple of (indata, outdata) with indata and outdata as numpy array
    data = list(data) if type(data) == tuple else data
    NumOfData = (set(self.inshape) ^ set(data[0].shape)).pop()
    if data[0].shape.index(NumOfData) != 0:
      x = list(range(0, len(data[0].shape)))
      x.pop(data[0].shape.index(NumOfData))
      data[0] = np.transpose(data[0], tuple([data[0].shape.index(NumOfData)]) + tuple(x))
    if data[1].shape.index(NumOfData) != 0:
      x = list(range(0, len(data[1].shape)))
      x.pop(data[1].shape.index(NumOfData))
      data[1] = np.transpose(data[1], tuple([data[1].shape.index(NumOfData)]) + tuple(x))

    # Training
    with alive_bar(NumOfGen) as bar:
      for _ in range (NumOfGen):
        Random = np.random.randint(0, NumOfData)
        input = data[0][Random]
        for i in range(len(self.model)):
          input = self.model[i].forward(input)

        outError = data[1][Random] - input
        for i in reversed(range(len(self.model))):
          outError = self.model[i].backward(outError)

        bar()

  def test(self, data):
    # data preprocessing: tuple of (indata, outdata) with indata and outdata as numpy array
    data = list(data) if type(data) == tuple else data
    NumOfData = (set(self.inshape) ^ set(data[0].shape)).pop()
    if data[0].shape.index(NumOfData) != 0:
      x = list(range(0, len(data[0].shape)))
      x.pop(data[0].shape.index(NumOfData))
      data[0] = np.transpose(data[0], tuple([data[0].shape.index(NumOfData)]) + tuple(x))
    if data[1].shape.index(NumOfData) != 0:
      x = list(range(0, len(data[1].shape)))
      x.pop(data[1].shape.index(NumOfData))
      data[1] = np.transpose(data[1], tuple([data[1].shape.index(NumOfData)]) + tuple(x))

    # Testing
    testcorrect = 0
    with alive_bar(NumOfData) as bar:
      for Generation in range (NumOfData):
        input = data[0][Generation]
        for i in range(len(self.model)):
          input = self.model[i].forward(input)

        testcorrect += 1 if np.argmax(input) == np.argmax(data[1][Generation]) else 0
        bar()

    return testcorrect / NumOfData


import numpy as np

class pool: 
  def __init__(self, stride, filtershape, opperation, inshape=None):
    self.stride, self.filtershape, self.opperation, self.inshape = stride, filtershape, opperation, inshape
    if inshape is not None and len(inshape) == 2:
      outshape = tuple([int(np.floor(inshape[0] / stride[0])), int(np.floor(inshape[1] / stride[1]))])
      while outshape[0] * stride[0] + filtershape[0] - stride[0] < inshape[0]:
        outshape[0] -= 1
      while outshape[1] * stride[1] + filtershape[1] - stride[1] < inshape[1]:
        outshape[1] -= 1
      self.outshape, self.out = outshape, np.zeros(outshape)
    elif inshape is not None and len(inshape) == 3:
      outshape = tuple([inshape[0], int(np.floor(inshape[1] / stride[0])), int(np.floor(inshape[2] / stride[1]))])
      while outshape[1] * stride[0] + filtershape[0] - stride[0] < inshape[1]:
        outshape[1] -= 1
      while outshape[2] * stride[1] + filtershape[1] - stride[1] < inshape[2]:
        outshape[2] -= 1
      self.outshape, self.out = outshape, np.zeros(outshape)

  def __copy__(self):
    return type(self)(self.stride, self.filtershape, self.opperation, self.inshape)

  def modelinit(self, inshape):
    self.inshape = inshape
    if len(inshape) == 2:
      outshape = tuple([int(np.floor(inshape[0] / self.stride[0])), int(np.floor(inshape[1] / self.stride[1]))])
      while outshape[0] * self.stride[0] + self.filtershape[0] - self.stride[0] < inshape[0]:
        outshape[0] -= 1
      while outshape[1] * self.stride[1] + self.filtershape[1] - self.stride[1] < inshape[1]:
        outshape[1] -= 1
    elif len(inshape) == 3:
      outshape = tuple([inshape[0], int(np.floor(inshape[1] / self.stride[0])), int(np.floor(inshape[2] / self.stride[1]))])
      while outshape[1] * self.stride[0] + self.filtershape[0] - self.stride[0] < inshape[1]:
        outshape[1] -= 1
      while outshape[2] * self.stride[1] + self.filtershape[1] - self.stride[1] < inshape[2]:
        outshape[2] -= 1
    self.outshape, self.out = outshape, np.zeros(outshape)
    return outshape

  def forward(self, input):
    self.input = input

    if self.opperation == 'Max':
      for i in range(self.outshape[1]):
        for j in range(self.outshape[2]):
          self.out[:, i, j] = np.amax(input[:, i * self.stride[0] : i * self.stride[0] + self.filtershape[0], j * self.stride[1] : j * self.stride[1] + self.filtershape[1]], axis=(1, 2))
    elif self.opperation == 'Min':
      for i in range(self.outshape[1]):
        for j in range(self.outshape[2]):
          self.out[:, i, j] = np.amin(input[:, i * self.stride[0] : i * self.stride[0] + self.filtershape[0], j * self.stride[1] : j * self.stride[1] + self.filtershape[1]], axis=(1, 2))
    elif self.opperation == 'Mean':
      for i in range(self.outshape[1]):
        for j in range(self.outshape[2]):
          self.out[:, i, j] = np.mean(input[:, i * self.stride[0] : i * self.stride[0] + self.filtershape[0], j * self.stride[1] : j * self.stride[1] + self.filtershape[1]], axis=(1, 2))
    elif self.opperation == 'Sum':
      for i in range(self.outshape[1]):
        for j in range(self.outshape[2]):
          self.out[:, i, j] = np.sum(input[:, i * self.stride[0] : i * self.stride[0] + self.filtershape[0], j * self.stride[1] : j * self.stride[1] + self.filtershape[1]], axis=(1, 2))
    return self.out

  def backward(self, outError):
    if self.opperation == 'Max' or self.opperation == 'Min':
      return np.repeat(np.repeat(outError, 2, axis=1), 2, axis=2) * np.equal(np.repeat(np.repeat(self.out, 2, axis=1), 2, axis=2), self.input).astype(int)
    elif self.opperation == 'Mean' or self.opperation == 'Sum':
      return np.repeat(np.repeat(outError, 2, axis=1), 2, axis=2) * np.repeat(np.repeat(self.out, 2, axis=1), 2, axis=2)


import numpy as np
from .activation import *

class gru:
  def __init__(self, outshape, outactivation, hidshape=64, learningrate=0.05, inshape=None):
    self.outactivation, self.learningrate = outactivation, learningrate
    self.inshape, self.hidshape, self.outshape = inshape, hidshape, outshape

    if inshape is not None:
      self.weightsinToResetGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
      self.weightsinToUpdateGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
      self.weightsinToHidGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))

    self.weightsHidToResetGate = np.random.uniform(-0.005, 0.005, (hidshape, hidshape))
    self.weightsHidToUpdateGate = np.random.uniform(-0.005, 0.005, (hidshape, hidshape))
    self.weightsHidToHidGate = np.random.uniform(-0.005, 0.005, (hidshape, hidshape))

    self.HidGatebias = np.zeros(hidshape)
    self.ResetGatebias = np.zeros(hidshape)
    self.UpdateGatebias = np.zeros(hidshape)

    self.weightsHidToout = np.random.uniform(-0.005, 0.005, (hidshape, outshape))
    self.outbias = np.zeros(outshape)

  def __copy__(self):
    return type(self)(self.outshape, self.outactivation, self.hidshape, self.learningrate, self.inshape)

  def modelinit(self, inshape):
    self.inshape = inshape
    self.weightsinToResetGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
    self.weightsinToUpdateGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
    self.weightsinToHidGate = np.random.uniform(-0.005, 0.005, (inshape, hidshape))
    return self.outshape

  def forward(self, input):
    self.input = input
    self.cellSize = len(input)

    self.Hid = np.zeros((self.cellSize + 1, self.hidshape))
    self.out = np.zeros((self.cellSize, self.outshape))
  
    self.ResetGate = np.zeros((self.cellSize, self.hidshape))
    self.UpdateGate = np.zeros((self.cellSize, self.hidshape))
    self.HidGate = np.zeros((self.cellSize, self.hidshape))

    for i in range(self.cellSize):
      self.ResetGate[i, :] = sigmoid().forward(self.weightsinToResetGate.T @ self.input[i, :] + self.weightsHidToResetGate.T @ self.Hid[i, :] + self.ResetGatebias)
      self.UpdateGate[i, :] = sigmoid().forward(self.weightsinToUpdateGate.T @ self.input[i, :] + self.weightsHidToUpdateGate.T @ self.Hid[i, :] + self.UpdateGatebias)
      self.HidGate[i, :] = tanh().forward(self.weightsinToHidGate.T @ self.input[i, :] + self.weightsHidToHidGate.T @ (self.Hid[i, :] * self.ResetGate[i, :]) + self.HidGatebias)
  
      self.Hid[i + 1, :] = (1 - self.UpdateGate[i, :]) * self.Hid[i, :] + self.UpdateGate[i, :] * self.HidGate[i, :]
      self.out[i, :] = self.outactivation.forward(self.weightsHidToout.T @ self.Hid[i + 1, :] + self.outbias)

    return self.out

  def backward(self, outError):
    inError = np.zeros(self.input.shape)
    HidError = np.zeros(self.hidshape)

    weightsinToResetGateΔ = np.zeros(self.weightsinToResetGate.shape)
    weightsinToUpdateGateΔ = np.zeros(self.weightsinToUpdateGate.shape)
    weightsinToHidGateΔ = np.zeros(self.weightsinToHidGate.shape)

    weightsHidToResetGateΔ = np.zeros(self.weightsHidToResetGate.shape)
    weightsHidToUpdateGateΔ = np.zeros(self.weightsHidToUpdateGate.shape)
    weightsHidToHidGateΔ = np.zeros(self.weightsHidToHidGate.shape)

    HidGatebiasΔ = np.zeros(self.HidGatebias.shape)
    ResetGatebiasΔ = np.zeros(self.ResetGatebias.shape)
    UpdateGatebiasΔ = np.zeros(self.UpdateGatebias.shape)

    weightsHidTooutΔ = np.zeros(self.weightsHidToout.shape)
    outbiasΔ = np.zeros(self.outbias.shape)

    for i in reversed(range(self.cellSize)):
      outGradient = self.outactivation.backward(self.out[i, :]) * outError[i, :]

      HidError += self.weightsHidToout @ outError[i, :]

      HidGateError = HidError * self.UpdateGate[i, :]
      UpdateGateError = HidError * (-1 * self.Hid[i, :]) + HidError * self.HidGate[i, :]
      ResetGateError = (self.weightsHidToHidGate.T @ HidGateError) * self.Hid[i, :]

      HidError += (self.weightsHidToHidGate.T @ HidGateError) * self.ResetGate[i, :] + self.weightsHidToResetGate.T @ ResetGateError + self.weightsHidToUpdateGate.T @ UpdateGateError
      inError[i, :] = self.weightsinToResetGate @ ResetGateError + self.weightsinToUpdateGate @ UpdateGateError + self.weightsinToHidGate @ HidGateError

      ResetGateGradient = sigmoid().backward(self.ResetGate[i, :]) * ResetGateError
      UpdateGateGradient = sigmoid().backward(self.UpdateGate[i, :]) * UpdateGateError
      HidGateGradient = tanh().backward(self.HidGate[i, :]) * HidGateError

      weightsinToResetGateΔ += np.outer(self.input[i, :].T, ResetGateGradient)
      weightsinToUpdateGateΔ += np.outer(self.input[i, :].T, UpdateGateGradient)
      weightsinToHidGateΔ += np.outer(self.input[i, :].T, HidGateGradient)
  
      weightsHidToResetGateΔ += np.outer(self.Hid[i, :].T, ResetGateGradient)
      weightsHidToUpdateGateΔ += np.outer(self.Hid[i, :].T, UpdateGateGradient)
      weightsHidToHidGateΔ += np.outer(self.Hid[i, :].T, HidGateGradient)
  
      HidGatebiasΔ += ResetGateGradient
      ResetGatebiasΔ += UpdateGateGradient
      UpdateGatebiasΔ += HidGateGradient

      weightsHidTooutΔ += np.outer(self.Hid[i].T, outGradient)
      outbiasΔ += outGradient

    self.weightsinToResetGate += np.clip(weightsinToResetGateΔ, -1, 1) * self.learningrate
    self.weightsinToUpdateGate += np.clip(weightsinToUpdateGateΔ, -1, 1) * self.learningrate
    self.weightsinToHidGate += np.clip(weightsinToHidGateΔ, -1, 1) * self.learningrate

    self.weightsHidToResetGate += np.clip(weightsHidToResetGateΔ, -1, 1) * self.learningrate
    self.weightsHidToUpdateGate += np.clip(weightsHidToUpdateGateΔ, -1, 1) * self.learningrate
    self.weightsHidToHidGate += np.clip(weightsHidToHidGateΔ, -1, 1) * self.learningrate

    self.HidGatebias += np.clip(HidGatebiasΔ, -1, 1) * self.learningrate
    self.ResetGatebias += np.clip(ResetGatebiasΔ, -1, 1) * self.learningrate
    self.UpdateGatebias += np.clip(UpdateGatebiasΔ, -1, 1) * self.learningrate

    self.weightsHidToout += np.clip(weightsHidTooutΔ, -1, 1) * self.learningrate
    self.outbias += np.clip(outbiasΔ, -1, 1) * self.learningrate

    return inError